
 /*******************************************************************/
 /*          S A S    A U T O  ---  Macro Library                   */
 /*                                                                 */
 /*    NAME: EQAR                                                   */
 /*   TITLE: Vector Autoregressive Process for PROC MODEL           */
 /* PRODUCT: ETS                                                    */
 /*    KEYS: ARIMA, Vector Autoregression                           */
 /*                                                                 */
 /* SUPPORT: Mark Little                 DATE:   02APR85            */
 /*                                      UPDATE: 06SEP88            */
 /*                                      UPDATE: 12JUL89            */
 /*                                                                 */
 /*******************************************************************/

 /*-------------------------------------------------------------------
  *
  *        Macro %EQAR to generate PROC MODEL code for autoregressive
  *        and vector autoregressive error processes.
  *
  * USAGE:
  *
  *        There are two cases:
  *
  *     %EQAR( name, nlag [, endolist [, laglist] ] [,M=] [,TYPE=] )
  *
  *       name     : prefix for %EQAR to use in constructing names
  *                  of variables needed to define the AR process.
  *
  *                  If the 'endolist' is not specified then the
  *                  endogenous list defaults to 'name' and 'name'
  *                  must be the name of the equation to which the
  *                  AR error process is to be applied.
  *
  *                  'name' can not exceed 32 characters.
  *
  *       nlag     : the order of the AR process.
  *
  *       enodlist : the list of equations to which the AR process
  *                  is to be applied.
  *
  *                  If more than one name is given an unrestricted
  *                  vector process is created with the structural
  *                  residuals of all the equations included as
  *                  regressors in each of the equations.
  *
  *                  If not specified, 'endolist' defaults to 'name'.
  *
  *       laglist  : the list of lags at which the AR terms are to be
  *                  added. The coefficients of the terms at lags not
  *                  listed are set to zero.
  *
  *                  All of the listed lags must be less than or equal
  *                  to 'nlag', and there must be no duplicates.
  *
  *                  If not specified, the 'laglist' defaults to
  *                  all lags 1 through 'nlag'.
  *
  *       M=method : specifies the estimation method to implement.
  *                  The values of M= are: M=CLS, M=ULS, ML=ML.
  *                  M=CLS: conditional least squares estimates.
  *                  M=ULS: unconditional least squares estimates.
  *                  ML=ML: maximum likelihood estimates.
  *
  *                  M=CLS is the default.
  *
  *                  Only M=CLS is allowed when more than one equation
  *                  is specified. The ULS and ML methods are not
  *                  supported for vector AR models by %EQAR.
  *
  *       TYPE=V   : specifies that the AR process is to be applied
  *                  to the endogenous variables themselves instead
  *                  of to the structural residuals of the equations.
  *
  *
  *
  *        An alternative usage of %EQAR is allowed to impose
  *        restrictions on a vector AR process by calling %EQAR
  *        several times to specify different AR terms and lags
  *        for different equations.
  *
  *        The first call has the form:
  *
  *     %EQAR( name, nlag, endolist, DEFER )
  *
  *       name     : prefix for %EQAR to use in constructing names
  *                  of variables needed to define the VAR process.
  *
  *       nlag     : the order of the AR process.
  *
  *       enodlist : the list of equations to which the AR process
  *                  is to be applied.
  *
  *       DEFER    : a special keyword that tells %EQAR to not generate
  *                  the AR process but to wait for further information
  *                  specified in later %EQAR calls for the same 'name'.
  *
  *        The later calls have the form:
  *
  *     %EQAR( name, eqlist [, varlist]  [, laglist] [,TYPE=] )
  *
  *       name     : same as in the first call.
  *
  *       eqlist   : the list of equations to which the specifications
  *                  in this %EQAR call are to be applied.
  *
  *                  Only names specified in the 'endolist' of the
  *                  first call for the 'name' can appear in 'eqlist'.
  *
  *       varlist  : the list of equations whose lagged structural
  *                  residuals are to be included as regressors in
  *                  the 'eqlist' equations.
  *
  *                  Only names specified in the 'endolist' of the
  *                  first call for 'name' can appear in 'varlist'.
  *
  *                  If not specified 'varlist' defaults to 'endolist'.
  *
  *       laglist  : the list of lags at which the AR terms are to be
  *                  added. The coefficients of the terms at lags not
  *                  listed are set to zero.
  *
  *                  All of the listed lags must be less than or equal
  *                  to 'nlag', and there must be no duplicates.
  *
  *                  If not specified, the 'laglist' defaults to
  *                  all lags 1 through 'nlag'.
  *
  *
  *
  * NOTES:
  *
  *        1. The calls to %EQAR must come AFTER all of the
  *           equations that the process applies to.
  *
  *        2. Use a short 'name' for the process if parameter
  *           estimates are to be written to an output data set.
  *           %EQAR tries to construct parameter names <= 32 characters,
  *           but this is limited by the length of 'name', which
  *           is used as a prefix for the ar parameter names.
  *
  *        3. See the SAS/ETS Guide, Version 6 Edition.
  *
  *
  * EXAMPLES:
  *              PROC MODEL DATA=A;
  *                PARMS A B C;
  *                Y = A + B*X1 + C*X2;
  *                %EQAR(Y,3)
  *                FIT Y;
  *
  *              PROC MODEL DATA=A;
  *                PARMS Y1_INT Y1_X1 Y2_INT Y2_X2;
  *                Y1 = Y1_INT + Y1_X1 * X1;
  *                Y2 = Y2_INT + Y2_X2 * X2;
  *                %EQAR(VEC_AR,2,Y1 Y2)
  *                FIT Y1 Y2;
  *
  *
  *
  *-----------------------------------------------------------------*/

 %MACRO EQAR( ARG1, ARG2, ARG3, ARG4, M=, TYPE= );

 %LOCAL PNAME NLAG ENDOLIST NENDO METHOD NLAGZLAG LAGZLAG TYPECODE;
 %LOCAL EQLIST NEQLIST VARLIST NVARLIST LAGLIST NLAGLIST HATNAME;
 %LOCAL CALLTYPE VAR VAR2 PARMNAME PARMLIST;
 %LOCAL I J K II JJ KK L L2 L3 N S1 S2;

      %LET PNAME = %UNQUOTE(%STR(&ARG1));

      %IF %LENGTH(&PNAME) <= 0 | %LENGTH(&PNAME) > 32 %THEN %DO;
            %PUT ERROR: Invalid AR process name "&PNAME".;
            %GOTO EXIT;
            %END;

      %GOTO DEFSUBS;
      %XRETURN:%DO;%END;

 %GLOBAL &PNAME;

      %IF %INDEX(0123456789,%SUBSTR(%QUOTE(&ARG2),1,1))>0 %THEN %DO;
         %LET CALLTYPE=D;

         %LET NLAG = %EVAL(&ARG2);
         %IF &NLAG<=0 | &NLAG>99 %THEN %DO;
            %PUT ERROR: Invalid lag length "&NLAG" for AR "&PNAME".;
            %GOTO EXIT;
            %END;
         %LET ENDOLIST = &ARG3;
         %IF %LENGTH(%QUOTE(&ENDOLIST))=0 %THEN %DO;
            %LET ENDOLIST = &PNAME; %LET NENDO = 1; %END;
         %ELSE %LET NENDO = %_COUNT_(&ENDOLIST);
         %DO I = 1 %TO &NENDO;
            %LET VAR = %SCAN(%QUOTE(&ENDOLIST),&I,%QUOTE( ));
            %IF %LENGTH(%QUOTE(&VAR))>40 %THEN %DO;
              %PUT ERROR: Invalid variable name "&VAR" for AR "&PNAME".;
               %LET NENDO = -999;
               %END;
            %END;
         %IF &NENDO<=0 | &NENDO>99 %THEN %DO;
            %PUT ERROR: Invalid variable list for AR "&PNAME".;
            %GOTO EXIT;
            %END;

         %IF %UPCASE(%QUOTE(&ARG4))^=DEFER %THEN %DO;
            %LET CALLTYPE=B;
            %LET EQLIST = &ENDOLIST;
            %LET VARLIST = &ENDOLIST;
            %LET NEQLIST = &NENDO;
            %LET NVARLIST = &NENDO;
            %_LLIST_ %IF &NLAGLIST<=0 %THEN %GOTO EXIT;
            %END;

         %ELSE
         %LET &PNAME = &NLAG,&NENDO,&ENDOLIST;

         %END;

      %ELSE %DO;
         %LET CALLTYPE=U;

         %LET L = %LENGTH(%QUOTE(&&&PNAME));
         %IF &L<=0 %THEN %DO;
            %PUT ERROR: The AR "&PNAME" is not defined.;
            %GOTO EXIT;
            %END;

         %LET NENDO = %SCAN(%QUOTE(&&&PNAME),2,%QUOTE(,));
         %LET ENDOLIST = %SCAN(%QUOTE(&&&PNAME),3,%QUOTE(,));
         %LET EQLIST = &ARG2;
         %IF %LENGTH(%QUOTE(&EQLIST))=0 %THEN %DO;
            %PUT ERROR: No equation list for AR "&PNAME".;
            %GOTO EXIT;
            %END;
         %LET NEQLIST = %_COUNT_(&EQLIST);
         %IF &NEQLIST>&NENDO %THEN %DO;
           %PUT ERROR: Too many items in equation list for AR "&PNAME".;
            %GOTO EXIT;
            %END;
         %IF &NEQLIST<=&NENDO %THEN
         %DO I = 1 %TO &NEQLIST;
            %LET VAR = %SCAN(%QUOTE(&EQLIST),&I);
            %IF %INDEX(&ENDOLIST,&VAR)=0 %THEN %DO;
               %PUT ERROR: Variable "&VAR" not in list for AR "&PNAME".;
               %LET NEQLIST = -1;
               %END;
            %END;
         %IF &NEQLIST<=0 %THEN %GOTO EXIT;

         %LET VARLIST = &ARG3;
         %IF %LENGTH(%QUOTE(&VARLIST))=0 %THEN %DO;
            %LET VARLIST = &ENDOLIST;
            %LET NVARLIST = &NENDO;
            %END;
         %ELSE %DO;
            %LET NVARLIST = %_COUNT_(&VARLIST);
            %IF &NVARLIST>&NENDO %THEN %DO;
           %PUT ERROR: Too many items in variable list for AR "&PNAME".;
               %GOTO EXIT;
               %END;
            %DO I = 1 %TO &NVARLIST;
               %LET VAR = %SCAN(%QUOTE(&VARLIST),&I);
               %IF %INDEX(&ENDOLIST,&VAR)=0 %THEN %DO;
                  %PUT ERROR: "&VAR" not in list for AR "&PNAME".;
                  %LET NVARLIST = -1;
                  %END;
               %END;
            %IF &NVARLIST<=0 %THEN %GOTO EXIT;
            %END;

         %LET NLAG = %SCAN(%QUOTE(&&&PNAME),1,%QUOTE(,));
         %_LLIST_ %IF &NLAGLIST<=0 %THEN %GOTO EXIT;

         %END;

 /* %*-- Get the TYPE and METHOD options------; */
      %IF %LENGTH(%QUOTE(&TYPE))>0 %THEN %DO;
         %LET TYPECODE = %SUBSTR(%UPCASE(&TYPE),1,1);
         %IF &TYPECODE^=E %THEN %DO;
            %PUT ERROR: Invalid TYPE=&TYPE for EQAR "&PNAME".;
            %GOTO EXIT;
            %END;
         %END;
      %ELSE %LET TYPECODE=E;

      %IF &TYPECODE=E %THEN %LET METHOD=CLS;
      %ELSE                 %LET METHOD=;

      %IF %LENGTH(%QUOTE(&M))>0 %THEN %DO;
         %IF &TYPECODE=V %THEN
         %PUT WARNING: M= option invalid for TYPE=V in AR "&PNAME".;
         %ELSE %DO;
            %IF %LENGTH(%QUOTE(&M))<=3 %THEN %DO;
               %LET METHOD = %UPCASE(&M);
               %LET NLAGZLAG = 0;
               %END;
            %ELSE %DO;
               %LET METHOD = %SUBSTR(%UPCASE(&M),1,3);
               %LET NLAGZLAG = %SUBSTR(%UPCASE(&M),4);
               %IF &NLAGZLAG=P %THEN %LET LAGZLAG = &NLAG;
               %IF &NLAGZLAG < 0 | &NLAGZLAG > %EVAL(2*&NLAG+20)
               %THEN %DO;
                  %PUT ERROR: Invalid numeric value in M=&M option;
                  %LET NLAGZLAG=0;
                  %END;
               %IF &METHOD^=CLS %THEN %DO;
                  %PUT ERROR: Invalid numeric value in M=&M option;
                  %LET NLAGZLAG=0;
                  %END;
               %END;
            %IF &METHOD^=ULS & &METHOD^=CLS & &METHOD^=ML
            %THEN %DO;
           %PUT ERROR: Invalid method M=&M for AR "&PNAME". M=CLS used.;
               %LET METHOD=CLS;
               %END;
            %END;
         %END;

      %IF &NENDO>1 %THEN
      %IF &METHOD=ULS | &METHOD=ML %THEN %DO;
 %PUT ERROR: Method M=&M not supported for vector process. M=CLS used.;
         %LET METHOD=CLS;
         %END;

 /*--- Determine best form for construction of parameter names ------*/
      %LET L  = %LENGTH(&PNAME);
      %LET L2 = %LENGTH(&NLAG);
      %LET L3 = %LENGTH(&NENDO);
      %LET S2 =;
      %IF &NENDO>1 %THEN
      %IF       &L2=1 & &L3=1 %THEN %DO;
         %IF       &L<=25 %THEN %LET S1 =_l;
         %ELSE %IF &L=26 %THEN %LET S1 =_;
         %ELSE %IF &L=27 %THEN %LET S1 =;
         %ELSE %DO;  %LET S1 =_l; /* Give up below 32*/
               %END;
         %END;
      %ELSE %IF &L2=2 & &L3=1 %THEN %DO;
         %IF       &L<=24 %THEN %LET S1 =_l;
         %ELSE  %IF  &L=25 %THEN %LET S1 =_;
         %ELSE %IF &L=26 %THEN %LET S1 =;
         %ELSE %DO;  %LET S1 =_l; /* Give up -------*/
               %END;
         %END;
      %ELSE %IF &L2=1 & &L3=2 %THEN %DO;
         %IF       &L=25 %THEN %DO;
             %LET S1 =;          /* a weak try */
             %END;
         %ELSE                %LET S1 =_l;
         %END;
      %ELSE                         %DO;
         %IF       &L<29 %THEN %DO; %LET S1 =_l; %LET S2 =_; %END;
         %ELSE %IF &L=29 %THEN %DO; %LET S1 =l;  %LET S2 =_; %END;
         %ELSE %IF &L=30 %THEN %DO; %LET S1 =;   %LET S2 =_; %END;
         %ELSE  %DO;
            %LET S1 =_l; %LET S2 =_;  /* Give up keeping below 32*/
            %END;
         %END;
      %ELSE
      %IF       &L2=1 %THEN %DO;
         %IF       &L<30 %THEN %LET S1 =_l;
         %ELSE %IF &L=30 %THEN %LET S1 =_;
         %ELSE                %LET S1 =_l;
         %END;
      %ELSE                 %DO;
         %IF       &L<29 %THEN %LET S1 =_l;
         %ELSE %IF &L=29 %THEN %LET S1 =_;
         %ELSE                %LET S1 =_l;
         %END;

 /*-----------------------------------------------------------------*/

      %IF &NENDO > 1 %THEN %LET HATNAME = &PNAME;
      %ELSE                %LET HATNAME = _RES_;

 /* %*-- Generate needed code -----; */

      %IF &CALLTYPE=D | &CALLTYPE=B %THEN %DO;

       %*-- insert assingment statement for temporary PRED. vars---;
         %IF &TYPECODE=E %THEN
         %DO I = 1 %TO &NENDO;
            %LET VAR = %SCAN(%QUOTE(&ENDOLIST),&I,%QUOTE( ));
        /*  &HATNAME._&VAR = ZLAG0( PRED.&VAR ); */
            &HATNAME._&VAR = &VAR;
            %END;

         %END;


      %IF &CALLTYPE=U | &CALLTYPE=B %THEN %DO;

         %DO I = 1 %TO &NEQLIST;

    /* %*-- insert assingment statement for endogenous variable---; */
            %LET VAR = %SCAN(%QUOTE(&EQLIST),&I,%QUOTE( ));
            %IF &NEQLIST=&NENDO %THEN %LET II = &I;
            %ELSE %DO JJ = 1 %TO &NENDO;
               %IF &VAR=%SCAN(%QUOTE(&ENDOLIST),&JJ,%QUOTE( )) %THEN %LET II=&JJ;
               %END;
            %LET PARMLIST = ;
            %LET N = 0;

            %IF &METHOD=ULS | &METHOD=ML %THEN %IF &NLAG>1 %THEN %DO;
               IF _OBS_ > &NLAG THEN
               %END;

            &VAR = &HATNAME._&VAR

            %DO J = 1 %TO &NLAGLIST;
               %LET L = %SCAN(%QUOTE(&LAGLIST),&J);
               %IF &L <= &NLAGZLAG %THEN %LET LAGZLAG = LAG;
               %ELSE                     %LET LAGZLAG = ZLAG;
               %DO K = 1 %TO &NVARLIST;
                  %LET VAR2 = %SCAN(%QUOTE(&VARLIST),&K,%QUOTE( ));
                  %IF &NVARLIST=&NENDO %THEN %LET KK = &K;
                  %ELSE %DO JJ = 1 %TO &NENDO;
                     %IF &VAR2=%SCAN(%QUOTE(&ENDOLIST),&JJ,%QUOTE( )) %THEN
                     %LET KK=&JJ;
                     %END;
                  %LET N = %EVAL(&N+1);
                  %IF &NENDO>1 %THEN
                  %LET PARMNAME = &PNAME.&S1.&L.&S2._&II._&KK;
                  %ELSE
                  %LET PARMNAME = &PNAME.&S1.&L;
                  %LET PARMLIST = &PARMLIST &PARMNAME;
                  %IF &TYPECODE=E %THEN
                  %STR( - &PARMNAME*&LAGZLAG&L(&HATNAME._&VAR2) );
                  %ELSE
                  %STR( + &PARMNAME*&LAGZLAG&L(&VAR2) );
                  %END;
               %END;
           %IF &NLAGZLAG > &NLAG %THEN %DO;
               %IF &TYPECODE=E %THEN
               %STR( + 0*LAG&NLAGZLAG(&VAR2-&HATNAME._&VAR2) );
               %ELSE
               %STR( + 0*LAG&NLAGZLAG(&VAR2) );
              %END;
           %STR(;)

     /* %*-- insert startup code for method=ULS or methdo=ML ---; */
            %IF &METHOD=ULS | &METHOD=ML %THEN %DO;
               %IF &NLAG>1 %THEN %_ULS_;
               %ELSE %DO;
                  IF _OBS_ = 1 THEN IF _ITER_>0 THEN
                     IF ABS(&PARMNAME)>=1
                        THEN  PUT 'WARNING: AT ' _ITER_=
                           ' the %EQAR process ' "&PNAME" ' is unstable.'
                           / '         ' &PARMNAME= ;
                  IF _MODE_ = "ESTIMATE" THEN DO;
                  IF _OBS_=1 THEN
                  &VAR = &HATNAME._&VAR * SQRT(1-&PARMNAME*&PARMNAME);
                  %IF &METHOD=ML %THEN %DO;
                    IF _MODE_ ^= "ESTIMATE" && (_NUSED_ = 0 OR MISSING(_NUSED_)) THEN _NUSED_ =1;
                    IF _OBS_=1 THEN
                    &PNAME._WGT_ =
                             (1-&PARMNAME*&PARMNAME)**(-1/(2*_NUSED_));
                    ELSE &PNAME._WGT_ = ZLAG1(&PNAME._WGT_);
                    &VAR = &VAR * &PNAME._WGT_;
                    END;
                    %END;
                  %END;
               %END;

     /* %*-- insert parms statements ---; */
            parms
            %LET N = 0;
            %DO J = 1 %TO &NLAGLIST;
               %DO K = 1 %TO &NVARLIST;
                  %LET N = %EVAL(&N+1);
                  %LET PARMNAME = %SCAN(%QUOTE(&PARMLIST),&n);
                  &PARMNAME 0
                  %END;
               %END;
                 ;

     /* %*-- construct label statements ---; */
            %LET N = 0;
            %DO J = 1 %TO &NLAGLIST;
               %LET L = %SCAN(%QUOTE(&LAGLIST),&J);
               %DO K = 1 %TO &NVARLIST;
                  %LET N = %EVAL(&N+1);
                  %LET PARMNAME = %SCAN(%QUOTE(&PARMLIST),&N);
                  %LET VAR2 = %SCAN(%QUOTE(&VARLIST),&K,%QUOTE( ));
                  label &PARMNAME =
                  %IF &NENDO>1 %THEN %DO;
                     %IF %LENGTH(&PNAME)+%LENGTH(&VAR)
                         +%LENGTH(&VAR2)+%LENGTH(&K) > 20
                     %THEN
                     %STR( "EQAR(&PNAME) &VAR: L&L,&VAR2"; );
                     %ELSE
                     %STR( "EQAR(&PNAME) &VAR: LAG&L parameter for &VAR2"; );
                     %END;
                  %ELSE %STR( "EQAR(&PNAME) &VAR lag&L parameter"; );
                  %END;
               %END;
            %END;

         %END;


      %GOTO ENDIT;

 /* %*--------------------------------------------------------------; */

 %DEFSUBS:%DO;%END;

 %MACRO _COUNT_( ARG, DEL );
      %LOCAL I N D;
      %IF %LENGTH(&DEL)=0 %THEN %LET D=%QUOTE( ); %ELSE %LET D=&DEL;
      %LET N=0;
      %DO I = 1 %TO 999;
         %IF %LENGTH(%SCAN(%QUOTE(&ARG),&I,&D))=0 %THEN %LET I = 9999;
         %ELSE %LET N = %EVAL(&N+1);
         %END;
         &N
    %MEND _COUNT_;


 %MACRO _LLIST_; %LOCAL I J;

      %LET LAGLIST = %UPCASE(&ARG4);
      %IF %LENGTH(&LAGLIST)>0 & %QUOTE(&LAGLIST)^=DEFER %THEN %DO;
         %LET NLAGLIST = %_COUNT_(&LAGLIST);
         %DO I = 1 %TO &NLAGLIST;
            %LET J = %SCAN(%QUOTE(&LAGLIST),&I);
            %IF &J<=0 | &J>&NLAG %THEN %DO;
               %PUT ERROR: Invalid lag value "&J" for "&PNAME".;
               %LET NLAGLIST = -999;
               %END;
            %END;
         %END;
      %ELSE %DO;
         %LET NLAGLIST = &NLAG;
         %LET LAGLIST = ;
         %DO I = 1 %TO &NLAGLIST;
            %LET LAGLIST = &LAGLIST &I;
            %END;
         %END;

      %IF &NLAGLIST<=0 | &NLAGLIST>&NLAG %THEN %DO;
         %PUT ERROR: Invalid lag list for AR "&PNAME".;
         %LET NLAGLIST = -1;
         %END;

    %MEND _LLIST_;

 /* %*--------------------------------------------------------------; */

 /* %*---MACRO TO GENERATE CODE FOR ULS AND ML METHODS-------------; */
 %MACRO _ULS_;

 %LOCAL P K P_1 K_1 P_K J N PARMS PARM;

 %MACRO _GAMMA_;

   %* GET PHI_J_J FROM REVERSED LEVINSON RECURRENCE;
      %DO K = 1 %TO &NLAG;
         &PNAME._PHI_&K = -%SCAN(%QUOTE(&PARMS),&K,?);
         %END;

      %IF &NLAG=2 %THEN %DO;
         &PNAME._PHI_1 = (&PNAME._PHI_1-&PNAME._PHI_2*&PNAME._PHI_1)
                        / (1 - &PNAME._PHI_2*&PNAME._PHI_2);
         %END;

      %ELSE
      %DO P = &NLAG %TO 2 %BY -1;
         %LET P_1 = %EVAL(&P-1);
         &PNAME._TEMP_1 = 1 / (1 - &PNAME._PHI_&P*&PNAME._PHI_&P);

         %DO K = 1 %TO &P_1;
            %LET P_K = %EVAL(&P-&K);
            &PNAME._L_&K = &PNAME._TEMP_1
               * (&PNAME._PHI_&K - &PNAME._PHI_&P*&PNAME._PHI_&P_K);
            %END;
         %DO K = 1 %TO &P_1; &PNAME._PHI_&K = &PNAME._L_&K; %END;

         %END;

      IF _ITER_>0 THEN
      IF ABS(&PNAME._PHI_1)>=1
      %DO K=2 %TO &NLAG; | ABS(&PNAME._PHI_&K)>=1 %END;
      THEN  PUT 'WARNING: AT ' _ITER_=
                ' THE %EQAR PROCESS ' "&PNAME" ' IS UNSTABLE.'
              / '         ' %DO K=1 %TO &NLAG; &PNAME._PHI_&K= %END; ;

      &PNAME._GAM_0 = 1
      %DO K=1 %TO &NLAG;/ (1 - &PNAME._PHI_&K*&PNAME._PHI_&K) %END; ;

      %IF &METHOD=ML %THEN %DO;
         &PNAME._WGT_ = &PNAME._GAM_0;
         %END;

      %DO P = 1 %TO &NLAG;  %LET P_1 = %EVAL(&P-1);

      %* GET PHI_P_K FROM FORWARD LEVINSON RECURRENCE;
         %IF &P=2 %THEN %DO;
            &PNAME._PHI_1 = &PNAME._PHI_1
                           + &PNAME._PHI_2*&PNAME._PHI_1;
            %END;
         %ELSE %IF &P>2 %THEN %DO;
            %DO K = 1 %TO &P_1;
               %LET P_K = %EVAL(&P-&K);
               &PNAME._L_&K = &PNAME._PHI_&K
                             + &PNAME._PHI_&P*&PNAME._PHI_&P_K;
               %END;
            %DO K = 1 %TO &P_1; &PNAME._PHI_&K = &PNAME._L_&K; %END;
            %END;

      %* COMPUTE GAMMA_P;
         &PNAME._GAM_&P =
         %DO K = 1 %TO &P; %LET P_K = %EVAL(&P-&K);
             - &PNAME._PHI_&K*&PNAME._GAM_&P_K
            %END; ;

      %* COMPUTE determinant-----;
         %IF &METHOD=ML %THEN %DO;
            &PNAME._WGT_ = &PNAME._WGT_ * ( &PNAME._GAM_0
            %DO K = 1 %TO &P;
               + &PNAME._PHI_&K*&PNAME._GAM_&K
               %END; );
            %END;

         %END;

      %IF &METHOD=ML %THEN %DO;
         &PNAME._WGT_ = &PNAME._WGT_ ** (1/(2*_NUSED_));
         %END;

    %MEND _GAMMA_;

 /* %*--------------------------------------------------------------; */

      %LET PARMS =;
      %LET N = 1;
      %DO P = 1 %TO &NLAGLIST;
         %LET K = %SCAN(%QUOTE(&LAGLIST),&P);
         %DO J = 1 %TO %EVAL(&K-&N);
            %LET PARMS =&PARMS?0; %LET N = %EVAL(&N+1); %END;
         %LET PARMS =&PARMS?%SCAN(%QUOTE(&PARMLIST),&P);
         %LET N = %EVAL(&N+1);
         %END;
      %DO J = &N %TO &NLAG;
         %LET PARMS =&PARMS?0; %LET N = %EVAL(&N+1); %END;
      %IF &N ^= %EVAL(&NLAG+1) %THEN
      %PUT ERROR: AR &PNAME ULS BUG IN SET UP OF PARMS. N=&N NLAG=&NLAG;

      IF _OBS_ <= &NLAG THEN DO;

         IF _OBS_ = 1 THEN DO; /*-- Initialize the Kalman filter --*/
            %_GAMMA_
            &PNAME._R_ = &PNAME._GAM_0;
            &PNAME._L_1 = &PNAME._GAM_1;
            &PNAME._K_1 = &PNAME._GAM_1;
            %DO K=1 %TO &NLAG; &PNAME._Z_&K = 0; %END;
            %DO K=2 %TO &NLAG;
            %LET K_2 = %EVAL(&K-2);
            &PNAME._L_&K = &PNAME._GAM_&K_2;
            &PNAME._K_&K = &PNAME._GAM_&K_2;
            %END;
            END;

         ELSE  DO; /*-- Update the Kalman filter --*/

         %* UDATE R;
            &PNAME._R_ = ZLAG1( &PNAME._R_
                        - &PNAME._L_1*&PNAME._L_1/&PNAME._R_ );

         %* UDATE THE ZS;
            %DO K=&NLAG %TO 2 %BY -1; %LET K_1 = %EVAL(&K-1);
               &PNAME._Z_&K = ZLAG1(
                       &PNAME._Z_&K_1+&PNAME._K_&K*&PNAME._TEMP_1 );
               %END;
            &PNAME._Z_1 = ZLAG1(
               %DO K=1 %TO &NLAG;
                  %LET PARM = %SCAN(%QUOTE(&PARMS),&K,?);
                  %IF %QUOTE(&PARM)^=0 %THEN
                  %DO;+&PARM*&PNAME._Z_&K %END;
                  %END;
                                      + &PNAME._K_1*&PNAME._TEMP_1 );

         %* UDATE THE KS;
            %DO K=&NLAG %TO 2 %BY -1; %LET K_1 = %EVAL(&K-1);
            &PNAME._K_&K = ZLAG1( &PNAME._K_&K
                                  - &PNAME._TEMP_2*&PNAME._L_&K_1 );
               %END;
            &PNAME._K_1 = ZLAG1(
                        &PNAME._K_1-&PNAME._TEMP_2*&PNAME._TEMP_3 );

         %* UDATE THE LS;
            %DO K=&NLAG %TO 2 %BY -1; %LET K_1 = %EVAL(&K-1);
            &PNAME._L_&K = ZLAG1( &PNAME._L_&K_1
                                     - &PNAME._TEMP_2*&PNAME._K_&K );
            %END;
            &PNAME._L_1 = ZLAG1( &PNAME._TEMP_3
                                      - &PNAME._TEMP_2*&PNAME._K_1 );
            END;

         &VAR = &HATNAME._&VAR + &PNAME._Z_1;
         &PNAME._TEMP_1 = - &VAR / &PNAME._R_;
         &VAR = &VAR / SQRT(&PNAME._R_);

         &PNAME._TEMP_2 = &PNAME._L_1 / &PNAME._R_;
         &PNAME._TEMP_3 =
         %DO K=1 %TO &NLAG;
            %LET PARM = %SCAN(%QUOTE(&PARMS),&K,?);
            %IF %QUOTE(&PARM)^=0 %THEN %DO;+&PARM*&PNAME._L_&K %END;
            %END; ;

         END;


      %IF &METHOD=ML %THEN %DO;
         IF _OBS_>1 THEN &PNAME._WGT_ = ZLAG1(&PNAME._WGT_);
         &VAR = &VAR * &PNAME._WGT_;
         %END;

    %MEND _ULS_;

    %GOTO XRETURN;

 /* %*--------------------------------------------------------------; */

 %EXIT:;
      ;ABORT; * %nrstr(%EQAR) ERROR INSERTED ABORT STATEMENT;

 %ENDIT:  %MEND EQAR;
